/*
 * kernel/power/pageflags.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * Suspend2 needs a few pageflags while working that aren't otherwise
 * used. To save the struct page pageflags, we dynamically allocate
 * a bitmap and use that. These are the only non order-0 allocations
 * we do.
 */
extern unsigned long * in_use_map;
extern unsigned long * pageset2_map;
extern unsigned long * checksum_map;
#ifdef CONFIG_DEBUG_PAGEALLOC
extern unsigned long * unmap_map;
#endif

#define PAGENUMBER(page) (page-mem_map)
#define PAGEINDEX(page) ((PAGENUMBER(page))/(8*sizeof(unsigned long)))
#define PAGEBIT(page) ((int) ((PAGENUMBER(page))%(8 * sizeof(unsigned long))))

/* 
 * freepagesmap is used in two ways: 
 * - During suspend, to tag pages which are not used (to speed up 
 *   count_data_pages);
 * - During resume, to tag pages which are in pagedir1. This does not tag 
 *   pagedir2 pages, so !== first use.
 */
#define PageInUse(page) \
	test_bit(PAGEBIT(page), &in_use_map[PAGEINDEX(page)])
#define SetPageInUse(page) \
	set_bit(PAGEBIT(page), &in_use_map[PAGEINDEX(page)])
#define ClearPageInUse(page) \
	clear_bit(PAGEBIT(page), &in_use_map[PAGEINDEX(page)])

#define PagePageset2(page) \
	(pageset2_map ? \
		test_bit(PAGEBIT(page), &pageset2_map[PAGEINDEX(page)]) : \
		0)
			
#define SetPagePageset2(page) \
	set_bit(PAGEBIT(page), &pageset2_map[PAGEINDEX(page)])
#define TestAndSetPagePageset2(page) \
	test_and_set_bit(PAGEBIT(page), &pageset2_map[PAGEINDEX(page)])
#define TestAndClearPagePageset2(page) \
	test_and_clear_bit(PAGEBIT(page), &pageset2_map[PAGEINDEX(page)])
#define ClearPagePageset2(page)	\
do { \
	if (pageset2_map) \
		clear_bit(PAGEBIT(page), &pageset2_map[PAGEINDEX(page)]); \
} while(0)

#define PageChecksumIgnore(page) \
	(checksum_map ? \
		test_bit(PAGEBIT(page), &checksum_map[PAGEINDEX(page)]) : \
		0)

#define SetPageChecksumIgnore(page) \
do { \
	if (checksum_map) \
		set_bit(PAGEBIT(page), &checksum_map[PAGEINDEX(page)]); \
} while(0)

#define ClearPageChecksumIgnore(page) \
do { \
	if (checksum_map) \
		clear_bit(PAGEBIT(page), &checksum_map[PAGEINDEX(page)]); \
} while(0)


#define SetUnmap(page) \
	set_bit(PAGEBIT(page), &unmap_map[PAGEINDEX(page)])
#define PageUnmap(page) \
	test_bit(PAGEBIT(page), &unmap_map[PAGEINDEX(page)])

extern int allocate_local_pageflags(unsigned long ** pagemap, int setnosave);
extern int free_local_pageflags(unsigned long ** pagemap);
extern void clear_map(unsigned long * pagemap);
