/*
 * kernel/power/proc.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It provides declarations for suspend to use in managing
 * /proc/software_suspend. When we switch to kobjects,
 * this will become redundant.
 *
 */

struct suspend_proc_data {
	char * filename;
	int permissions;
	int type;
	union {
		struct {
			unsigned long * bit_vector;
			int bit;
		} bit;
		struct {
			int * variable;
			int minimum;
			int maximum;
		} integer;
		struct {
			unsigned long * variable;
			unsigned long minimum;
			unsigned long maximum;
		} ul;
		struct {
			char * variable;
			int max_length;
		} string;
		struct {
			void * read_proc;
			void * write_proc;
			void * data;
		} special;
	} data;
	
	/* Side effects routines. Used, eg, for reparsing the
	 * resume2 entry when it changes */
	int (* read_proc) (void);
	int (* write_proc) (void); 
	struct list_head proc_data_list;
};

#define SUSPEND_PROC_DATA_CUSTOM	0
#define SUSPEND_PROC_DATA_BIT		1
#define SUSPEND_PROC_DATA_INTEGER	2
#define SUSPEND_PROC_DATA_UL		3
#define SUSPEND_PROC_DATA_STRING	4

#define PROC_WRITEONLY 0200
#define PROC_READONLY 0400
#define PROC_RW 0600

struct proc_dir_entry * suspend_register_procfile(
		struct suspend_proc_data * suspend_proc_data);
void suspend_unregister_procfile(struct suspend_proc_data * suspend_proc_data);

